package flm.b4a.cache;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;

import android.graphics.Bitmap;
import android.graphics.Bitmap.Config;
import android.graphics.BitmapFactory;
import android.os.StatFs;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.collections.List;
import anywheresoftware.b4a.objects.collections.Map;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper.BitmapWrapper;
import flm.b4a.cache.DiskLruCache.Editor;
import flm.b4a.cache.DiskLruCache.Snapshot;

@BA.Version(1.38F)
@BA.ShortName("Cache")
@BA.Author("Frdric Leneuf-Magaud")
@BA.Events(values = {
		"CopyDone(Key As String, Error As Boolean)",
		"PutDone(Key As String, Error As Boolean)",
		"GetDone(Key As String, Obj As Object)"
})
public class Cache
{
	private int bitmapCacheSize = 0;
	private int bufferSize = 0;
	private LruCache<String, Bitmap> bitmapCache;
	private DiskLruCache diskCache;
	private boolean UseMemCache; 

	/**
	 *Initializes the caches.
	 *<B>memoryCachePct</B>: percentage of free memory to use for the memory cache (this cache is only used for bitmaps, the other objects are stored on disk). Cannot exceed 90.
	 *<B>diskCacheSize</B>: size in bytes of the cache on disk.
	 *<B>diskCacheDir</B>: folder where the cache files are stored.
	 * If empty, the default cache folder is used. This folder is specific to the application.
	 *
	 *Example: <I>myCache.Initialize(25, 20 * 1024 * 1024, "/mnt/sdcard/myAppCache") </I> '25% for the memory cache and 20MB for the disk cache
	 */
	public void Initialize(BA ba, byte memoryCachePct, long diskCacheSize, String diskCacheDir) throws IOException
	{
		if (memoryCachePct > 0)
		{
			// Max size for the memory cache = 90% of the free memory
			bitmapCacheSize = (int) ((getFreeMemory() / 100) * Math.min(memoryCachePct, 90));
			bitmapCache = new LruCache<String, Bitmap>(bitmapCacheSize) {
				@Override
				protected int sizeOf(String key, Bitmap bitmap) {
					return bitmap.getRowBytes() * bitmap.getHeight();
				}
			};
			UseMemCache = (bitmapCache != null && bitmapCacheSize > 0);
		}
		else
			UseMemCache = false;
		if (diskCacheSize > 0)
		{
			// If no cache dir is specified, the default cache folder is used
			final File fCacheDir;
			if (diskCacheDir.length() == 0)
				fCacheDir = ba.context.getCacheDir();
			else
				fCacheDir = new File(diskCacheDir);
			fCacheDir.mkdirs();
			try {
				StatFs statfs = new StatFs(fCacheDir.toString());
				bufferSize = statfs.getBlockSize();
			}
			catch(Exception e) {
				BA.Log(e.getMessage());
				bufferSize = 8192;
			}
			diskCache = DiskLruCache.open(fCacheDir, 1, 1, diskCacheSize);
		}
	}

	/**
	 *Caches the given bitmap in memory.
	 *Returns false if the bitmap is too big for the cache.
	 */
	public boolean PutBitmapInMemory(String key, BitmapWrapper bitmap)
	{
		if (BitmapSize(bitmap) > bitmapCacheSize)
			return false;
		synchronized(bitmapCache) {
			bitmapCache.put(key, bitmap.getObject());
		}
		return true;
	}

	/**
	 *Returns the size in bytes of the given bitmap.
	 */
	public final long BitmapSize(BitmapWrapper bitmap)
	{
		return bitmap.getObject().getRowBytes() * bitmap.getObject().getHeight();
	}

	/**
	 *Returns the Bitmap for <I>key</I> if it exists in the cache.
	 *Returns null if not found.
	 */
	public final Bitmap GetBitmapFromMemory(String key)
	{
		synchronized(bitmapCache) {
			return bitmapCache.get(key);
		}
	}

	/**
	 *Removes the Bitmap for <I>key</I> if it exists.
	 */
	public final void RemoveBitmapFromMemory(String key)
	{
		synchronized(bitmapCache) {
			bitmapCache.remove(key);
		}
	}

	/**
	 *Clears the memory cache holding bitmaps.
	 */
	public final void ClearMemoryCache(BA ba) throws Exception
	{
		synchronized(bitmapCache) {
			try {
				bitmapCache.evictAll();
			}
			catch (Exception e) {
				// L'erreur est ignore si le cache a bien t nettoy
				if (bitmapCache.size() > 0)
					throw e;
			}
		}
	}

	/**
	 *Returns a copy of the current contents of the memory cache, ordered from least recently accessed to most recently accessed.
	 */
	public final Map MemoryList()
	{
		synchronized(bitmapCache) {
			Map m = new Map();
			m.Initialize();
			m.getObject().putAll(bitmapCache.snapshot());
			return m;
		}
	}

	/**
	 *Returns true if the Bitmap for <I>key</I> is still in the memory cache.
	 */
	public final boolean IsInMemory(String key)
	{
		synchronized(bitmapCache) {
			return bitmapCache.isInCache(key);
		}
	}

	/**
	 *Returns the sum of the sizes in bytes of the entries in the memory cache.
	 */
	public final int MemoryUsed()
	{
		synchronized(bitmapCache) {
			return bitmapCache.size();
		}
	}

	/**
	 *Returns the available space in bytes of the memory cache.
	 */
	public final int MemoryFree()
	{
		synchronized(bitmapCache) {
			return bitmapCacheSize - bitmapCache.size();
		}
	}

	/**
	 *Returns the number of bitmaps that have been evicted from the memory cache.
	 */
	public final int MemoryEvictionCount()
	{
		synchronized(bitmapCache) {
			return bitmapCache.evictionCount();
		}
	}

	/**
	 *Returns the number of times <I>GetBitmapFromMemory</I> returned a bitmap that was present in the cache.
	 */
	public final int MemoryHitCount()
	{
		synchronized(bitmapCache) {
			return bitmapCache.hitCount();
		}
	}

	/**
	 *Returns the number of times <I>GetBitmapFromMemory</I> returned null (bitmap not found).
	 */
	public final int MemoryMissCount()
	{
		synchronized(bitmapCache) {
			return bitmapCache.missCount();
		}
	}

	/**
	 *Returns the number of times <I>PutBitmapInMemory</I> was called.
	 */
	public final int MemoryPutCount()
	{
		synchronized(bitmapCache) {
			return bitmapCache.putCount();
		}
	}

	/**
	 *Returns the directory where the disk cache stores its data.
	 */
	public final String GetDirectory()
	{
		synchronized(diskCache) {
			return diskCache.getDirectory().toString();
		}
	}

	/**
	 *Returns the sum of the sizes in bytes of the files in the disk cache.
	 */
	public final long DiskUsed()
	{
		synchronized(diskCache) {
			return diskCache.size();
		}
	}

	/**
	 *Returns the available space in bytes of the disk cache.
	 */
	public final long DiskFree()
	{
		synchronized(diskCache) {
			return diskCache.getMaxSize() - diskCache.size();
		}
	}

	/**
	 *Returns true if the disk cache has been closed.
	 */
	public final boolean IsClosed() throws IOException
	{
		synchronized(diskCache) {
			return diskCache.isClosed();
		}
	}

	/**
	 *Closes the disk cache. Stored values will remain on the filesystem.
	 */
	public final void CloseDiskCache() throws IOException
	{
		synchronized(diskCache) {
			diskCache.close();
		}
	}

	/**
	 *Closes the disk cache and deletes all files in the cache directory including files that weren't created by the cache.
	 */
	public final void CloseAndClearDiskCache() throws IOException
	{
		synchronized(diskCache) {
			diskCache.delete();
		}
	}

	/**
	 *Deletes all files in the cache directory including files that weren't created by the cache.
	 */
	public final void ClearDiskCache() throws IOException
	{
		synchronized(diskCache) {
			diskCache.delete();
			final File fCacheDir = diskCache.getDirectory();
			final long diskCacheSize = diskCache.getMaxSize();
			diskCache = DiskLruCache.open(fCacheDir, 1, 1, diskCacheSize);
		}
	}

	/**
	 *Stores the given string in the disk cache.
	 */
	public final void PutStringOnDisk(String key, String value) throws IOException
	{
		synchronized(diskCache) {
			Editor editor = diskCache.edit(key);
			try {
				editor.set(0, value);
			}
			finally {
				editor.commit();
			}
		}
	}

	/**
	 *Gets the string for <I>key</I> from the disk cache.
	 *Returns an empty string if the key was not found.
	 */
	public final synchronized String GetStringFromDisk(String key) throws IOException
	{
		Snapshot snapshot = diskCache.get(key);
		if (snapshot == null)
			return "";

		String result = "";
		try {
			result = snapshot.getString(0);
		}
		finally {
			snapshot.close();
		}
		return result;
	}

	/**
	 *Removes the file for <I>key</I> if it exists in the disk cache.
	 *Returns true if the file was removed.
	 */
	public final boolean RemoveFromDisk(String key) throws IOException
	{
		synchronized(diskCache) {
			return diskCache.remove(key);
		}
	}

	/**
	 *Returns true if the file for <I>key</I> is still in the disk cache.
	 */
	public final boolean IsOnDisk(String key)
	{
		return new File(GetDirectory(), key + "." + 0).exists();
	}

	/**
	 *Stores the given bitmap in the disk cache.
	 *<B>Format</B>: "JPEG" or "PNG"
	 *<B>Quality</B>: quality of JPEG compression (0..100). This setting is ignored for the PNG format.
	 */
	public final synchronized void PutBitmapOnDisk(String key, BitmapWrapper bitmap, String format, int quality) throws IOException
	{
		Editor editor = diskCache.edit(key);
		OutputStream out = null;
		try {
			out = new BufferedOutputStream(editor.newOutputStream(0), bufferSize);
			bitmap.getObject().compress(Bitmap.CompressFormat.valueOf(format.toUpperCase(BA.cul)), quality, out);
			out.flush();
		}
		finally {
			if (out != null)
				out.close();
			editor.commit();
		}
	}

	/**
	 *Stores the given bitmap in the disk cache (asynchronously).
	 *<B>Format</B>: "JPEG" or "PNG"
	 *<B>Quality</B>: quality of JPEG compression (0..100). This setting is ignored for the PNG format.
	 *<B>eventPrefix</B> = prefix of the "PutDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_PutDone(key As String, Error As Boolean)</I>
	 *<I>Error</I> is true if there was an error.
	 */
	public final void PutBitmapOnDisk_Async(final BA ba, final String key, final BitmapWrapper bitmap, final String format, final int quality, final String eventPrefix)
	{
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					PutBitmapOnDisk(key, bitmap, format, quality);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase() + "_putdone", (eventPrefix != ""), new Object[]{key, false});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase() + "_putdone", (eventPrefix != ""), new Object[]{key, true});
				}
			}
		};
		BA.submitRunnable(r, null, 0);
	}

	/**
	 *Copies a file in the disk cache.
	 */
	public final synchronized void CopyFileToDisk(String key, String dir, String filename) throws IOException
	{
		Editor editor = diskCache.edit(key);
		OutputStream out = null;
		try {
			InputStream input;
			if (dir.toLowerCase().compareTo("assetsdir") == 0)
				input = BA.applicationContext.getAssets().open(filename);
			else
				input = new FileInputStream(new File(dir, filename));

			out = new BufferedOutputStream(editor.newOutputStream(0), bufferSize);
			byte[] buffer = new byte[bufferSize];
			int bytesRead = 0;
			while((bytesRead = input.read(buffer)) > 0) {
				out.write(buffer, 0, bytesRead);
			}
			out.flush();
		}
		finally {
			if (out != null)
				out.close();
			editor.commit();
		}
	}

	/**
	 *Copies a file in the disk cache (asynchronously).
	 *<B>eventPrefix</B> = prefix of the "CopyDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_CopyDone(key As String, Error As Boolean)</I>
	 *<I>Error</I> is true if there was an error.
	 */
	public final void CopyFileToDisk_Async(final BA ba, final String key, final String dir, final String filename, final String eventPrefix)
	{
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					CopyFileToDisk(key, dir, filename);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_copydone", (eventPrefix != ""), new Object[]{key, false});
				}
				catch (IOException e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_copydone", (eventPrefix != ""), new Object[]{key, true});
				}
			}
		};
		BA.submitRunnable(r, null, 0);
	}

	/**
	 *Stores the given object in the disk cache.
	 *This works only with serializable objects (numbers, booleans, lists, arrays...).
	 */
	public final synchronized void PutObjectOnDisk(String key, Object obj) throws IOException, NotSerializableException
	{
		Editor editor = diskCache.edit(key);
		OutputStream out = null;
		try {
			out = new BufferedOutputStream(editor.newOutputStream(0), bufferSize);

			ObjectOutputStream objos = new ObjectOutputStream(out);
			objos.writeObject(obj);
			objos.flush();

			out.flush();
		}
		finally {
			if (out != null)
				out.close();
			editor.commit();
		}
	}

	/**
	 *Stores the given object in the disk cache (asynchronously).
	 *This works only with serializable objects (numbers, booleans, lists, arrays...).
	 *<B>eventPrefix</B> = prefix of the "PutDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_PutDone(key As String, Error As Boolean)</I>
	 *<I>Error</I> is true if there was an error.
	 */
	public final void PutObjectOnDisk_Async(final BA ba, final String key, final Object obj, final String eventPrefix)
	{
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					PutObjectOnDisk(key, obj);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_putdone", (eventPrefix != ""), new Object[]{key, false});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_putdone", (eventPrefix != ""), new Object[]{key, true});
				}
			}
		};
		BA.submitRunnable(r, null, 0);
	}

	/**
	 *Returns the object for <I>key</I> from the disk cache.
	 *Returns null if the key was not found.
	 */
	public final synchronized Object GetObjectFromDisk(String key) throws IOException, ClassNotFoundException
	{
		Snapshot snapshot = diskCache.get(key);
		if (snapshot == null)
			return null;

		InputStream in = null;
		try {
			in = new BufferedInputStream(snapshot.getInputStream(0), bufferSize);
			ObjectInputStream objos = new ObjectInputStream(in);
			return objos.readObject();
		}
		finally {
			if (in != null)
				in.close();
			snapshot.close();
		}
	}

	/**
	 *Returns the object for <I>key</I> from the disk cache (asynchronously).
	 *<B>eventPrefix</B> = prefix of the "GetDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_GetDone(key As String, obj As Object)</I>
	 *<I>obj</I> is null if the key was not found or there was an error.
	 */
	public final void GetObjectFromDisk_Async(final BA ba, final String key, final String eventPrefix)
	{
		Runnable r = new Runnable() {
			@Override
			public void run() {
				Object o = null;
				try {
					o = GetObjectFromDisk(key);
				} catch (Exception e) {
					ba.setLastException(e);
				}
				ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_getdone", (eventPrefix != ""), new Object[]{key, o});
			}
		};
		BA.submitRunnable(r, null, 0);
	}

	/**
	 *Returns the bitmap for <I>key</I> from the disk cache.
	 *Returns null if the key was not found.
	 *<B>colorReduction</B> = if True, the number of colors is reduced to save memory. This setting has no effect if some pixels are not opaque.
	 */
	public final synchronized Bitmap GetBitmapFromDisk(String key, boolean colorReduction) throws Exception
	{
		Snapshot snapshot = diskCache.get(key);
		if (snapshot == null)
			return null;

		InputStream in = null;
		try {
			in = new BufferedInputStream(snapshot.getInputStream(0), bufferSize);
			BitmapFactory.Options opts = new BitmapFactory.Options();
			if (colorReduction) opts.inPreferredConfig = Config.RGB_565;
			Bitmap b = BitmapFactory.decodeStream(in, null, opts);
			if (b == null) {
				Exception e = new Exception("Error loading bitmap (GetBitmapFromDisk)");
				throw e;
			}
			return b;
		}
		catch (OutOfMemoryError OoME) {
			Exception e = new Exception("Not enough memory (GetBitmapFromDisk)");
			throw e;
		}
		finally {
			if (in != null)
				in.close();
			snapshot.close();
		}
	}

	/**
	 *Returns the bitmap for <I>key</I> from the disk cache (asynchronously).
	 *<B>colorReduction</B> = if True, the number of colors is reduced to save memory. This setting has no effect if some pixels are not opaque.
	 *<B>eventPrefix</B> = prefix of the "GetDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_GetDone(key As String, bmp As Bitmap)</I>
	 *<I>bmp</I> is null if the key was not found or there was an error.
	 */
	public final void GetBitmapFromDisk_Async(final BA ba, final String key, final boolean colorReduction, final String eventPrefix)
	{
		Runnable r = new Runnable() {
			@Override
			public void run() {
				Bitmap b = null;
				try {
					b = GetBitmapFromDisk(key, colorReduction);
				} catch (Exception e) {
					ba.setLastException(e);
				}
				ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_getdone", (eventPrefix != ""), new Object[]{key, b});
			}
		};
		BA.submitRunnable(r, null, 0);
	}

	/**
	 *Returns the bitmap for <I>key</I> from one of the caches (the memory cache is looked first).
	 *Returns null if the key was not found.
	 *The bitmap source (memory or disk cache) is specified in the unfiltered log.
	 *<B>updateMemory</B> = if true, the memory cache is updated if the bitmap was only on disk.
	 */
	public final synchronized Bitmap GetBitmap(String key, boolean updateMemory) throws Exception
	{
		if (UseMemCache && IsInMemory(key)) {
			System.out.println("From memory cache: " + key);			
			return GetBitmapFromMemory(key);
		}
		else {
			if (UseMemCache && updateMemory) {
				Bitmap b = GetBitmapFromDisk(key, false);
				System.out.println("From disk cache: " + key);			
				if (b != null) { 
					bitmapCache.put(key, b);
				}
				return b;
			}
			else
				return GetBitmapFromDisk(key, false);
		}
	}

	/**
	 *Returns the given bitmap for <I>key</I> from one of the caches.
	 *If the bitmap is in memory, it is sent back immediately. Otherwise, it is loaded from disk asynchronously.
	 *<B>updateMemory</B> = if true, the memory cache is updated if the bitmap was only found on disk.
	 *<B>eventPrefix</B> = prefix of the "GetDone" event.
	 *An event is triggered when the job ends. You can get it with:
	 *<I>Sub eventPrefix_GetDone(key As String, bmp As Bitmap)</I>
	 *<I>bmp</I> is null if the key was not found or there was an error.
	 */
	public final void GetBitmap_Async(final BA ba, final String key, final boolean updateMemory, final String eventPrefix) throws IOException
	{
		if (UseMemCache && IsInMemory(key)) {
			Bitmap bitmem = GetBitmapFromMemory(key);
			ba.raiseEvent2(this, false, eventPrefix.toLowerCase(BA.cul) + "_getdone", (eventPrefix != ""), new Object[]{key, bitmem});
		}
		else {
			Runnable r = new Runnable() {
				@Override
				public void run() {
					Bitmap b = null;
					try {
						b = GetBitmapFromDisk(key, false);
						if (UseMemCache && updateMemory && b != null) { 
							bitmapCache.put(key, b);
						}
					} catch (Exception e) {
						ba.setLastException(e);
					}
					ba.raiseEventFromDifferentThread(this, null, 0, eventPrefix.toLowerCase(BA.cul) + "_getdone", (eventPrefix != ""), new Object[]{key, b});
				}
			};
			BA.submitRunnable(r, null, 0);
		}
	}

	/**
	 *Returns a list of the current contents of the disk cache, unordered.
	 *Map structure: key=filename, value=object array (0=time of the last modification, 1=size in bytes) 
	 *
	 *Example:
	 *<I>Dim m As Map
	 *m = myCache.DiskList
	 *Dim Info(2) As Object
	 *For i = 0 To m.Size - 1
	 *	Log(m.GetKeyAt(i))
	 *	Info = m.GetValueAt(i)
	 *	Log("   " &amp; DateTime.Date(Info(0)) &amp; " " &amp; DateTime.Time(Info(0)))
	 *	Log("   " &amp; Info(1) &amp; " bytes")
	 *Next</I>
	 */
	public final synchronized Map DiskList()
	{
		Map m = new Map();
		m.Initialize();
		File cacheDir = diskCache.getDirectory();
		File journal = new File(cacheDir, DiskLruCache.JOURNAL_FILE);
		for (File f: cacheDir.listFiles()) {
			if (f.toString().compareTo(journal.toString()) != 0)
			{
				Object Info[] = new Object[2];
				Info[0] = f.lastModified();
				Info[1] = f.length();
				m.Put(f.toString(), Info);
			}
		}
		return m;
	}

	/**
	 *Returns a copy of the journal of the disk cache.
	 *
	 *Example:
	 *<I>Dim l As List
	 *l = myCache.DiskJournal
	 *For i = 0 To l.Size - 1
	 *	Log(l.Get(i))
	 *Next</I>
	 */
	public final synchronized List DiskJournal() throws IOException
	{
		List l = new List();
		l.Initialize();

		File Journal = new File(diskCache.getDirectory(), DiskLruCache.JOURNAL_FILE);
		BufferedReader reader = new BufferedReader(new FileReader(Journal));
		String line;
		while ((line = reader.readLine()) != null) {
			l.Add(line);
		}
		reader.close();

		return l;
	}

	/**
	 *Gets/sets the buffer size of the disk cache. By default, it is set to the block size of the filesystem. 
	 */
	public int getBufferSize()
	{
		return bufferSize;
	}
	public synchronized void setBufferSize(int Size)
	{
		bufferSize = Size;
	}

	/**
	 *Gets the maximum amount of memory (in bytes) that may be used by the running program.
	 */
	public static long getMaxMemory()
	{
		Runtime r = java.lang.Runtime.getRuntime();
		return r.maxMemory();
	}

	/**
	 *Gets an approximation of the amount (in bytes) of free memory available to the running program.
	 */
	public static long getFreeMemory()
	{
		Runtime r = java.lang.Runtime.getRuntime();
		final long MM = r.maxMemory();		//maximum amount of memory that may be used by the running program
		final long TM = r.totalMemory();	//current heap size
		final long FM = r.freeMemory();		//free heap memory
		if (android.os.Build.VERSION.SDK_INT < 11)
		{
			// Before Honeycomb
			final long NA = android.os.Debug.getNativeHeapAllocatedSize();
			return Math.max(0, MM - TM + FM - NA);
		}
		else
			return Math.max(0, MM - TM + FM);
	}
}
